
from .biocham import import_biocham
from .cadbiom import import_cadbiom
from .sbgnpd import import_sbgnpd

CONVERTERS = {
    "biocham": import_biocham,
    "cadbiom": import_cadbiom,
    "sbgnpd": import_sbgnpd,
}






import sys

from xml.dom import *
from xml.dom.minidom import parse

ENTITY_CLASSES = [
        "simple chemical",
        "macromolecule",
        "macromolecule multimer",
        "nucleic acid feature",
        "perturbing agent",
        "unspecified entity",
        "complex",
    ]
PROCESS_CLASSES = [
        "process",
        "association",
        "dissociation",
        "uncertain process",
        "omitted process",
    ]
LOGIC_CLASSES = [
        "and",
        "or",
        "not",
    ]
EMPTY_CLASSES = [
        "source and sink",
    ]
COMPARTMENT_CLASSES = [
        "compartment",
    ]
IGNORE_CLASSES = [
    ]
SUPPORTED_GLYPHS =  \
    ENTITY_CLASSES + \
    PROCESS_CLASSES + \
    LOGIC_CLASSES + \
    EMPTY_CLASSES + \
    COMPARTMENT_CLASSES + \
    IGNORE_CLASSES

MODIFICATION_ARCS = [
    "production",
    "consumption",
]
MODULATION_ARCS = [
    "catalysis",
    "stimulation",
    "necessary stimulation",
    "inhibition"
]
LOGIC_ARCS = [
    "logic arc",
]
SUPPORTED_ARCS = \
    MODIFICATION_ARCS + \
    MODULATION_ARCS + \
    LOGIC_ARCS


type2abbrv = {
    "simple chemical": "s",
    "macromolecule": "m",
    "macromolecule multimer": "mm",
    "complex": "cplx",
    "source and sink": "void",
    "process": "p",
    "association": "a",
    "dissociation": "d",
    "uncertain process": "up",
    "omitted process": "op",
}
def get_type_abbrv(t):
    abbrv = type2abbrv[t]
    return abbrv

molecule2abbrv = {
    "macromolecule multimer": "m_",
}
def get_molecule_abbrv(t):
    abbrv = molecule2abbrv.get(t, "")
    return abbrv

class SbgnGlyph:
    compartmentRef = None
    compartment = None
    def __init__(self, dom):
        self.dom = dom
        self.id = self.dom.getAttribute("id")
        self.type = self.dom.getAttribute("class")
        self.ports = {}
        for n in self.dom.childNodes:
            if n.nodeName == "port":
                self.ports[n.getAttribute("id")] = self.id
        if self.dom.hasAttribute("compartmentRef"):
            self.compartmentRef = self.dom.getAttribute("compartmentRef")

    def set_compartment(self, cg):
        self.compartment = cg

    def __repr__(self):
        return "%s(%s,%s)" % (self.__class__.__name__,self.type, self.id)

    def __lt__(g1, g2):
        return repr(g1) < repr(g2)

def attr_of_glyph(n):
    cls = n.getAttribute("class")
    assert cls in ["state variable", "unit of information"]
    if cls == "state variable":
        e, a = "state", "value"
    else:
        e, a = "label", "text"
    n = n.getElementsByTagName(e)
    if n:
        n = n[0]
        label = n.getAttribute(a) or "_"
        if cls == "state variable" and n.hasAttribute("variable"):
            label += "@%s" % n.getAttribute("variable")
        return label

class SbgnCompartment(SbgnGlyph):
    def __init__(self, *args):
        SbgnGlyph.__init__(self, *args)
        for n in self.dom.childNodes:
            if n.nodeName == "label":
                self.label = n.getAttribute("text")

class SbgnLogic(SbgnGlyph):
    def __init__(self, *args):
        SbgnGlyph.__init__(self, *args)
        self.inputs = []
    def add_input(self, e):
        self.inputs.append(e)


class SbgnEntity(SbgnGlyph):
    def __init__(self, *args):
        SbgnGlyph.__init__(self, *args)
        self.consumers = []
        self.producers = []
        self.__is_clone = len(self.dom.getElementsByTagName("clone")) > 0

    def __hash__(self):
        return hash(self.name)

    def __lt__(e1, e2):
        if e1.type in ENTITY_CLASSES \
            and e2.type in ENTITY_CLASSES:
            t_e1 = ENTITY_CLASSES.index(e1.type)
            t_e2 = ENTITY_CLASSES.index(e2.type)
            if t_e1 != t_e2:
                return t_e1 < t_e2
            if hasattr(e1, "__lt_sametype__"):
                return e1.__lt_sametype__(e2)
        return SbgnGlyph.__lt__(e1,e2)

    @property
    def name(self):
        return "%s_%s" % (get_type_abbrv(self.type), self.id)

    def is_clone(self):
        return self.__is_clone
    def is_void(self):
        return self.type in EMPTY_CLASSES

    def register_consumer(self, p):
        self.consumers.append(p)
    def register_producer(self, p):
        self.producers.append(p)


class SbgnMolecule(SbgnEntity):
    def __init__(self, dom):
        SbgnEntity.__init__(self, dom)
        self.attrs = []
        for n in self.dom.childNodes:
            if n.nodeName == "label":
                self.label = n.getAttribute("text")
            elif n.nodeName == "glyph":
                attr = attr_of_glyph(n)
                if attr:
                    self.attrs.append(attr)

    def __lt_sametype__(m1, m2):
        if m1.label != m2.label:
            return m1.label < m2.label
        def r(l):
            if l[0] == '_':
                return " %s" % l[1:]
            return l
        return tuple(map(r, m1.attrs)) < tuple(map(r, m2.attrs))

    @property
    def name(self):
        name = "%s%s" % (get_molecule_abbrv(self.type), self.label)
        if self.attrs:
            name += "-"+"".join(self.attrs)
        if self.compartment:
            name += "/%s" % self.compartment.label
        return name


class SbgnComplex(SbgnEntity):
    def __init__(self, dom):
        SbgnEntity.__init__(self, dom)
        self.attrs = []
        self.components = []
        for n in self.dom.childNodes:
            if n.nodeName == "label":
                self.label = n.getAttribute("text")
            elif n.nodeName == "glyph":
                cls = n.getAttribute("class")
                if cls == "complex":
                    self.components.append(SbgnComplex(n))
                elif cls in ENTITY_CLASSES:
                    self.components.append(SbgnMolecule(n))
                else:
                    attr = attr_of_glyph(n)
                    if attr:
                        self.attrs.append(attr)

    @property
    def name(self):
        if self.components:
            name = "c("+":".join(sorted([c.name for c in self.components]))+")"
        else:
            assert self.label, "Unsupported complex glyph (%s)"%self.dom.toxml()
            name = self.label
        if self.attrs:
            name += "-"+"".join(self.attrs)
        if self.compartment:
            name += "/%s" % self.compartment.label
        return name



class SbgnProcess(SbgnGlyph):
    def __init__(self, *args):
        SbgnGlyph.__init__(self, *args)
        self.name = "%s_%s" % (get_type_abbrv(self.type), self.id)
        self.productions = []
        self.consumptions = []
        self.modulations = []

    def is_frontier(self):
        return len(self.consumptions) == 0 \
                or len(self.productions) == 0

    def register_production(self, e):
        if not e.is_void():
            self.productions.append(e)
            e.register_producer(self)

    def register_consumption(self, e):
        if not e.is_void():
            self.consumptions.append(e)
            e.register_consumer(self)

    def register_modulation(self, cls, e):
        self.modulations.append((cls, e))



class SbgnPDModel:
    def __init__(self, dom):
        root = dom.documentElement
        maps = root.getElementsByTagName("map")
        assert len(maps) == 1
        assert maps[0].getAttribute("language") == "process description"
        self.mainmap = maps[0]

        self.entities = {}
        self.processes = {}
        self.logic = {}
        self.ports = {}
        self.compartments = {}
        self.name2entity = {}

        for n in self.mainmap.childNodes:
            if n.nodeName == "glyph":
                cls = n.getAttribute("class")
                assert cls in SUPPORTED_GLYPHS, "Unknown glyph class '%s'" % cls
                g = None
                if cls == "complex":
                    g = SbgnComplex(n)
                    r = self.entities
                elif cls in ENTITY_CLASSES:
                    g = SbgnMolecule(n)
                    r = self.entities
                elif cls in EMPTY_CLASSES:
                    g = SbgnEntity(n)
                    r = self.entities
                elif cls in PROCESS_CLASSES:
                    g = SbgnProcess(n)
                    r = self.processes
                elif cls in LOGIC_CLASSES:
                    g = SbgnLogic(n)
                    r = self.logic
                elif cls in COMPARTMENT_CLASSES:
                    g = SbgnCompartment(n)
                    r = self.compartments
                else:
                    assert cls in IGNORE_CLASSES, cls
                if g is not None:
                    r[g.id] = g
                    self.ports.update(g.ports)
                    if g.compartmentRef is not None:
                        g.set_compartment(self.compartments[g.compartmentRef])
                    if hasattr(g, "name"):
                        self.register_entity(g)

            elif n.nodeName == "arc":
                cls = n.getAttribute("class")
                assert cls in SUPPORTED_ARCS, "Unknown arc class '%s'" % cls

                def resolve(id):
                    id = self.resolve_id(id)
                    if id in self.entities:
                        return self.resolve_clone(self.entities[id])
                    elif id in self.processes:
                        return self.processes[id]
                    else:
                        return self.logic[id]

                source = resolve(n.getAttribute("source"))
                target = resolve(n.getAttribute("target"))
                    
                if cls == "production":
                    source.register_production(target)
                elif cls == "consumption":
                    try:
                        target.register_consumption(source)
                    except:
                        print("Error for arc %s: source is %s" % (n.getAttribute("id"), repr(target)), file=sys.stderr)
                        raise
                elif cls in MODULATION_ARCS:
                    target.register_modulation(cls, source)
                elif cls in LOGIC_ARCS:
                    target.add_input(source)
                else:
                    assert False, cls

    def register_entity(self, e):
        name = e.name
        if name in self.name2entity:
            self.name2entity[name].append(e)
        else:
            self.name2entity[name] = [e]

    def resolve_id(self, gid):
        if gid in self.ports:
            return self.ports[gid]
        return gid

    def resolve_clone(self, ce):
        if ce.is_clone():
            clones = []
            for e in self.name2entity[ce.name]:
                if e.is_clone():
                    clones.append((e.id, e))
            return min(clones)[1]
        else:
            return ce


def parse_sbgnpd(fd):
    dom = parse(fd)
    return SbgnPDModel(dom)





from functools import partial

from .export_utils import pint_of_localtransition

try:
    import boolean
except ImportError:
    print("ERROR: Please install boolean.py package (pip3 install boolean.py)")
    raise

class BoolToAN:
    def __init__(self, ba, ls_of_lit, out, subs=None):
        self.ba = ba
        self.ls_of_lit = ls_of_lit
        self.out = out
        self.subs = subs

    @staticmethod
    def BooleanAlgebra():
        return boolean.BooleanAlgebra()

    def make_transitions(self, changes, expr):
        LT = partial(pint_of_localtransition, changes)

        dnf = self.ba.dnf(expr)
        if self.subs:
            # substitution can break the DNF form
            expr = dnf.subs(self.subs)
            dnf = self.ba.dnf(expr)

        if isinstance(dnf, self.ba.OR):
            clauses = dnf.args
        elif isinstance(dnf, self.ba.AND):
            clauses = [dnf]
        else:
            clauses = [dnf]
        for clause in clauses:
            if isinstance(clause, self.ba.AND):
                lits = clause.args
            elif clause == self.ba.TRUE:
                lits = None
            else:
                lits = [clause]
            conds = [self.ls_of_lit(l) for l in lits] if lits else []
            self.out(LT(conds))





from functools import reduce

from .lib.sbgnpd import *
from .lib.export_utils import pint_of_ls

def import_sbgnpd(sbgnpd_filename, outfd, initial_state=(), stories=(),
                    iface=None, fancy_names=False):
    from .lib.boolean_utils import BoolToAN

    ba = BoolToAN.BooleanAlgebra()

    model = parse_sbgnpd(sbgnpd_filename)
    def resolve_ids(ids):
        return [model.resolve_clone(model.entities[eid]) for eid in ids]

    # format stories specification
    stories_spec = []
    for i, story in enumerate(stories):
        if isinstance(story, tuple) and len(story) == 2 \
                and isinstance(story[0], str) \
                and isinstance(story[1], (tuple, list, set, frozenset)):
            spec = {"name": story[0], "ids": tuple(story[1])}
        else:
            spec = {"name": "story{}".format(i), "ids": story}
        spec["entities"] = resolve_ids(spec["ids"])
        spec["sid"] = i
        stories_spec.append(spec)
    stories = stories_spec

    def story_name(sid):
        return stories[sid]["name"]

    e2stories = {}
    for i, story in enumerate(stories):
        for e in story["entities"]:
            if e not in e2stories:
                e2stories[e] = [i]
            else:
                e2stories[e].append(i)

    if fancy_names:
        def an_name(e):
            return e.name
    else:
        def an_name(e):
            return e.id

    # declare conflicts
    conflicts = {}
    for story in stories:
        procs = set()
        for e in story["entities"]:
            procs.update(e.consumers)
        for p in procs:
            if p not in conflicts:
                conflicts[p] = set()
            conflicts[p].update(procs)

    # filter self-conflicts
    for p in conflicts:
        if p in conflicts[p]:
            conflicts[p].remove(p)

    # declare automata for entities
    # TODO? clones
    entities = set([e for e in model.entities.values() if not e.is_void()])

    name_clash = {}
    for e in entities:
        name = an_name(e)
        assert name not in name_clash, \
            "Name clash [%s] between %s and %s" % (name, e, name_clash[name])
        name_clash[name] = e

    def local_states_of_entity(e, present=True):
        sids = e2stories.get(e)
        if sids:
            ret = []
            for sid in sids:
                a = story_name(sid)
                if present:
                    ret.append((a,an_name(e)))
                else:
                    ret += [(a,an_name(f)) for f in stories[sid]["entities"] if f.id != e.id] \
                        + [(a, 0)]
            return ret
        else:
            return [(an_name(e), 1 if present else 0)]

    if iface is not None:
        iface["resolve_ids"] = resolve_ids
        iface["id2stories"] = {}
        iface["id2ls"] = {}
        iface["automaton_to_entity"] = {}
        for e in entities:
            iface["id2stories"][e.id] = e2stories.get(e, ())
            iface["id2ls"][e.id] = local_states_of_entity(e)

    def _p(name):
        return '"{}"'.format(name)
    def out(data):
        print(data, file=outfd)

    for e in entities:
        if e not in e2stories:
            out("{} [0, 1]".format(_p(an_name(e))))
            if iface is not None:
                iface["automaton_to_entity"][an_name(e)] = e
    for story in stories:
        out("{} [{}]".format(_p(story["name"]),
                        ", ".join(["0"] + [_p(an_name(e)) for e in story["entities"]])))

    class StoryState:
        def __init__(self, sid, i):
            self.sid = sid
            self.i = i
        def __repr__(self):
            return "%s(%s,%s)" % (self.__class__.__name__,self.sid, self.i)
        def __lt__(g1, g2):
            return repr(g1) < repr(g2)

    class LS:
        def __init__(self, a, i):
            self.a = a
            self.i = i
            self.si = str(i)
        def __lt__(ai, bj):
            return (ai.a,ai.si) < (bj.a,bj.si)
        def to_pint(self):
            return (self.a, self.i)

    def expand_lit(b, pos):
        if isinstance(b, StoryState): # story,0
            assert pos
            return ba.symbols(LS(story_name(b.sid), b.i))[0]
        else:
            local_states = [LS(a,i) for a,i in local_states_of_entity(b, pos)]
            symbols = ba.symbols(*local_states)
            return reduce(lambda x,y: x|y, symbols)

    def ls_of_lit(lit):
        assert not isinstance(lit, ba.NOT)
        return lit.obj.to_pint()

    subs = {}

    def Lit(e):
        x = ba.symbols(e)[0]
        if x not in subs:
            subs[x] = expand_lit(e, True)
            if not isinstance(e, StoryState):
                subs[~x] = expand_lit(e, False)
        return x

    def logic_from_entity(n):
        if n.type in ENTITY_CLASSES:
            return Lit(n)
        elif n.type == "and":
            expr = logic_from_entity(n.inputs[0])
            for m in n.inputs[1:]:
                expr &= logic_from_entity(m)
            return expr
        elif n.type == "or":
            expr = logic_from_entity(n.inputs[0])
            for m in n.inputs[1:]:
                expr |= logic_from_entity(m)
            return expr

    def logic_of_modulations(modulations):
        expr = ba.FALSE
        necessary = ba.TRUE
        for (cls, var) in modulations:
            varexpr = logic_from_entity(var)
            if cls == "necessary stimulation":
                necessary &= varexpr
            elif cls in ["stimulation", "catalysis","modulation"]:
                expr |= varexpr
            elif cls in ["inhibition", "modulation"]:
                expr |= ~varexpr
        if expr is ba.FALSE:
            expr = ba.TRUE
        return necessary & expr


    out_trs = []

    b2a = BoolToAN(ba, ls_of_lit, out_trs.append, subs)

    #
    # transitions + automata for explicit processes
    #
    for p in model.processes.values():
        conds = logic_of_modulations(p.modulations)
        pconflicts = conflicts.get(p, ())

        if not pconflicts and len(p.consumptions) == 0:
            for e in p.productions:
                for a,j in local_states_of_entity(e):
                    b2a.make_transitions([(a,0,j)], conds)
        elif not pconflicts and len(p.productions) == 0 and len(p.consumptions) == 1:
            e = p.consumptions[0]
            trs = [(a,i,0) for a,i in local_states_of_entity(e)]
            b2a.make_transitions(trs, conds)
        else:

            implicit_process = not pconflicts and \
                                len(p.consumptions) <= 1 and \
                                len(p.productions) <= 1
            if implicit_process:
                for e in p.consumptions + p.productions:
                    if e not in e2stories:
                        implicit_process = False
                        break

            if not implicit_process:
                out("{} [0,1]".format(an_name(p)))
                for e in p.consumptions:
                    conds &= Lit(e)
                for c in pconflicts:
                    conds &= ~Lit(c)
                b2a.make_transitions([(an_name(p),0,1)], conds)
                conds = Lit(p)

            gstories = {}
            def push_sids(sids):
                for sid1 in sids:
                    for sid2 in sids:
                        if sid1 not in gstories:
                            gstories[sid1] = set()
                        gstories[sid1].add(sid2)
            def fill_group(group, sid):
                if sid in group:
                    return
                group.add(sid)
                sids = gstories[sid]
                for sid in sids:
                    fill_group(group, sid)

            consumptions = p.consumptions
            productions = p.productions

            inv = set(p.productions).intersection(p.consumptions)
            if inv:
                consumptions = set(consumptions) - inv
                productions = set(productions) - inv

            cons_indiv_e = []
            cons_s2e = {}
            prod_s2e = {}
            prod_done = ba.TRUE
            for e in consumptions:
                sids = e2stories.get(e)
                if not sids:
                    cons_indiv_e.append(e)
                else:
                    push_sids(sids)
                    for sid in sids:
                        cons_s2e[sid] = e
            for e in productions:
                sids = e2stories.get(e)
                if not sids:
                    b2a.make_transitions([(an_name(e),0,1)], conds)
                else:
                    push_sids(sids)
                    for sid in sids:
                        prod_s2e[sid] = e
                prod_done &= Lit(e)

            known = set()
            groups = []
            for sid in gstories:
                if sid in known:
                    continue
                group = set()
                fill_group(group, sid)
                groups.append(group)
                known.update(group)

            for sids in groups:
                changes = []
                for sid in sids:
                    a = story_name(sid)
                    i = cons_s2e.get(sid)
                    j = prod_s2e.get(sid)
                    i = an_name(i) if i is not None else 0
                    j = an_name(j) if j is not None else 0
                    changes.append((a,i,j))
                    if not j:
                        prod_done &= Lit(StoryState(sid,0))
                b2a.make_transitions(changes, conds)

            # process de-activation
            if not implicit_process:
                done = prod_done
                b2a.make_transitions([(an_name(p), 1, 0)], done)
            else:
                done = ba.TRUE

            for e in cons_indiv_e:
                b2a.make_transitions([(an_name(e),1,0)], conds & done)

    for spec in sorted(out_trs):
        out(spec)

    if initial_state:
        init = []
        initial_state = resolve_ids(initial_state)
        for e in initial_state:
            init += local_states_of_entity(e)
        out("initial_state {}".format(", ".join(map(pint_of_ls, init))))





import os
import sys

from .cfg import *

if IN_IPYTHON:
    from colomoto_jupyter import disp_jupyter_js
    from IPython.display import display, Markdown, HTML

from colomoto_jupyter.ui import *

def enable_dbg():
    """
    Enable debug output (executed command lines, etc.)
    """
    CFG["dbg"] = True
    os.environ["OCAMLRUNPARAM"] = "b"
    if IN_IPYTHON:
        disp_jupyter_js("pypint_jsapi.debug_enabled(true);")

def disable_dbg():
    """
    Disable debug output (executed command lines, etc.)
    """
    CFG["dbg"] = False
    del os.environ["OCAMLRUNPARAM"]
    if IN_IPYTHON:
        disp_jupyter_js("pypint_jsapi.debug_enabled(false);")

def dbg(msg):
    """
    If debug is enabled, print `msg` (prefixed with '#')
    """
    if CFG["dbg"]:
        if IN_IPYTHON:
            print("# %s" % msg)
        else:
            print("# %s" % msg, file=sys.stderr)

if IN_IPYTHON:
    def info(msg):
        """
        Display `msg` (interpreted in Markdown format)
        """
        display(Markdown(msg))
else:
    def info(msg):
        """
        Print `msg` (note: in IPython, `msg` is interpreted in Markdown
        format)
        """
        print(msg, file=sys.stderr)

__all__ = [
    "dbg",
    "enable_dbg",
    "disable_dbg",
    "info"
]





import os

def file_ext(filename):
    """
    Returns the extension of `filename` in lower case; or None if there is no
    extension
    """
    filename = os.path.basename(filename)
    if "." in filename:
        return filename.split(".")[-1].lower()






import json
import re

re_int = re.compile(r"\d+")

class Inconc:
    """
    Inconclusive type (in opposition to `True` and `False`).
    It rises a `TypeError` when used in a Boolean expression.
    """
    def __init__(self):
        raise TypeError("Inconc should not be instantiated")
    def __bool__(self):
        raise TypeError("Inconc is not a Boolean")
    __nonzero__ = __bool__

def ternary(b):
    """
    Returns ternary variable representation of `b`: either `True`, `False`, or
    :py:class:`.Inconc`. If `b` is `None`, returns `Inconc`.
    """
    if b is None:
        return Inconc
    elif b in [True,False,Inconc]:
        return b
    else:
        raise TypeError("'%s' is not a valid ternary" % repr(b))

def pint_value(value):
    """
    Returns string representation of a value for Pint
    """
    if isinstance(value, int):
        return str(value)
    else:
        return "\"{}\"".format(value)

class Conditions(dict):
    """
    Used to represent conditions of transitions.
    They are modeled as standard dictionnary, associating automata to the
    required local states.
    """
    def __init__(self, *args):
        super(Conditions, self).__init__(*args)
    def __str__(self):
        """
        Pint text representation of a condition.
        """
        return " and ".join(["{}={}".format(*map(pint_value, it)) for it in self.items()])

class LocalTransition(object):
    """
    A local transition within a single automaton.

    An instance has 4 attributes:

    * `a`: name of the automaton
    * `i`: initial local state of `a`
    * `j`: final local state of `a` after the transition
    * `conds`: :py:class:`.Conditions` object
    """
    def __init__(self, a, i, j, conds):
        """
        Transition `a` `i` -> `j` when `conds`, where `conds` is a `dict`-like
        object.
        """
        self.a = a
        self.i = i
        self.j = j
        self.conds = Conditions(conds)

    def __repr__(self):
        """
        Pint text representation of a local transition
        """
        r_conds = " when %s" % self.conds if self.conds else ""
        return "{} {} -> {}{}".format(*map(pint_value,
                (self.a, self.i, self.j, r_conds)))

    @property
    def modified_automata(self):
        """
        Set of automata in which the local transition takes place
        (``[a]``)
        """
        return set([self.a])

    @property
    def origins(self):
        return frozenset([(self.a, self.i)])

class SynchronizedLocalTransitions(object):
    """
    Synchronized local transitions.

    An instance has 2 attributes:

    * `local_transitions`: a list of `(a,i,j)` tuples representing the local
      transitions acting (always) synchronously
    * `conds`: :py:class:`.Conditions` object
    """
    def __init__(self, aijs, conds):
        """
        `aijs` is a list of `(a,i,j)` tuples and `conds` a `dict`-like object.
        """
        assert len(aijs) > 1
        self.local_transitions = [tuple(aij) for aij in aijs]
        self.__automata = frozenset([a for (a,_,_) in aijs])
        self.conds = Conditions(conds)

    def __repr__(self):
        """
        Pint text representation of synchronized local transitions.
        """
        r_conds = " when %s" % self.conds if self.conds else ""
        return "{ %s }%s" % \
            (" ; ".join(["{} {} -> {}".format(*map(pint_value, aij)) \
                for aij in self.local_transitions]), r_conds)

    @property
    def modified_automata(self):
        """
        Set of automata in which the local transitions take place.
        """
        return set(self.__automata)

    @property
    def origins(self):
        return frozenset([(a,i) for (a,i,_) in self.local_transitions])


def local_transition_from_json(tup):
    """
    Converts a JSON representation of a local transition (as return by Pint
    executables) to either :py:class:`.LocalTransition` or
    :py:class:`.SynchronizedLocalTransitions`.
    """
    if len(tup) == 4:
        return LocalTransition(*tup)
    elif len(tup) == 2:
        return SynchronizedLocalTransitions(*tup)
    raise ValueError("%s: Invalid tuple for local transition" % tup)


class Goal:
    """
    Object which represents reachability goal used in most of Pint commands.

    In its simplest form, a goal is the local state of an automaton.
    A goal can also be a sequence of sub-states.

    It can be instanciated either with arguments or keywords (but not both).
    If instanciated with arguments, each argument specifies a sub-state, and the
    goal is the sequence of these sub-states.
    A sub-state can be specified either by a string (in pint format), or by a
    `dict`-like object.
    If instanciated with keywords, the goal is a single sub-state where the keys
    correspond to the automata and the values their local states.

    Finally, alternative goals can be constructed using the operator ``|`` on
    the instanciated individual goals (see :py:meth:`.Goal.__or__`).


    Examples:

    >>> Goal(a=1)          # simple goal
    >>> Goal("a=1")        # equivalent to above
    >>> Goal(a=1,b=1)      # single sub-state goal
    >>> Goal("a=1,b=1")    # equivalent to above
    >>> Goal("a=1", "b=1") # sequence of simple goals
    >>> Goal({"a": 1, "b": 1}, {"a": 0})    # sequence of sub-state goals
    >>> Goal(a=1) | Goal(b=1)  # alternative goals
    """
    def __init__(self, *args, **kwargs):
        if args and kwargs:
            raise TypeError("Goal cannot be instanciated with both arguments and keywords")
        if not args and not kwargs:
            raise TypeError("Goal is empty")

        def parse_ls(s):
            a,i = s.split("=")
            a = a.strip().strip('"')
            i = i.strip()
            if re_int.fullmatch(i):
                i = int(i)
            else:
                i = i.strip('"')
            return a,i

        def goal_of_arg(a):
            if isinstance(a, str):
                return dict(map(parse_ls, a.split(",")))
            elif isinstance(a, dict):
                return a.copy()
            else:
                raise TypeError("Goal: do not support %s as argument" % type(a))

        if args:
            if isinstance(args[0], Goal):
                self.__goals = []
                for arg in args:
                    if not isinstance(arg, Goal):
                        raise TypeError("Cannot mix Goal with other types in arguments")
                    self.__goals += arg.__goals
            else:
                self.__goals = [[goal_of_arg(arg) for arg in args]]
        if kwargs:
            self.__goals = [[kwargs]]

        self.__automata = set()
        for gs in self.__goals:
            for g in gs:
                self.__automata.update(g.keys())

    @classmethod
    def from_arg(self, arg, **kwargs):
        """
        Returns a `Goal` instance corresponding to `arg`
        """
        if isinstance(arg, self):
            return arg
        elif isinstance(arg, str) or isinstance(arg, dict):
            return self(arg)
        elif isinstance(arg, list):
            return self(*arg)
        elif isinstance(arg, tuple) and len(arg) == 2:
            return self({arg[0]: arg[1]})
        elif arg is None:
            return self(**kwargs)
        else:
            raise TypeError("Cannot convert a %s to %s" % (type(arg), self))

    def __or__(g1, g2):
        """
        Construction of *alternative* goals: the returned goal consists of the
        disjunction of `g1` and `g2` (at least one of the two goals have to be
        reached).

        :rtype: Goal
        """
        return Goal(g1, g2)

    @property
    def automata(self):
        """
        Returns the set of automata that are part of the goal specification
        """
        return self.__automata.copy()

    @property
    def is_simple_goal(self):
        """
        ``True`` iff the goal is only the local state of one automaton
        """
        return len(self.__goals) == 1 \
            and len(self.__goals[0]) == 1 \
            and len(self.__goals[0][0]) == 1 \

    def to_pint(self):
        """
        Returns the argument to append to Pint commands for goal specification
        """
        args = []
        for gs in self.__goals:
            if args:
                args.append("or")
            args += [",".join([pint_of_localstates([ai]) for ai in g.items()]) \
                        for g in gs]
        return " ".join(args)

    def __str__(self):
        return self.to_pint()


def pint_of_localstates(items):
    """
    Returns Pint text representation of a list of local states
    """
    def fmt_values(i):
        if type(i) is int:
            return [str(i)]
        elif type(i) is str:
            return ["\"%s\"" % i]
        else:
            return [fmt_values(j)[0] for j in i]
    def pint_of_keyvalue(a,i):
        return ["\"%s\"=%s" % (a,i) for i in fmt_values(i)]
    lss = []
    for a, i in items:
        lss += pint_of_keyvalue(a,i)
    return ",".join(lss)

__all__ = [
    "Inconc", "ternary",
    "Conditions", "LocalTransition", "SynchronizedLocalTransitions",
    "local_transition_from_json",
    "Goal",
    "pint_of_localstates",
]




import re
import sys

from xml.dom.minidom import parse

from pypint.cfg import CFG

"""
Unsupported (TODO):
- MacroNodes
- transition/action
- transition/event
- transition/fact_ids
- transition/CDATA
"""

def import_cadbiom(localfile, outfd, full_context=True, relax_threshold=6):
    try:
        import boolean
    except ImportError:
        print("ERROR: Please install boolean.py package (pip3 install boolean.py)")
        raise
    from .lib.boolean_utils import BoolToAN

    ba = boolean.BooleanAlgebra()

    bcx = parse(localfile)

    trMap = {}
    def tr(name):
        return trMap.get(name, name)

    def register_name(name) :
        if ord("0") <= ord(name[0]) <= ord("9"):
            new_name = "v%s" % name
            trMap[name] = new_name
            return new_name
        return name

    # fetch nodes
    SimpleNodes = []
    for nt in ["CSimpleNode", "CTrapNode", "CMacroNode", "CStartNode"]:
        SimpleNodes += [register_name(n.getAttribute("name")) for n in bcx.getElementsByTagName(nt)]
    PermNodes = [register_name(n.getAttribute("name")) for n in bcx.getElementsByTagName("CPermNode")]

    def declare_automaton(a, *states):
        print("%s [%s]" % (a, ",".join(map(str,states))), file=outfd)

    # declare automata
    for a in SimpleNodes + PermNodes:
        declare_automaton(a, 0, 1)

    def ls_of_lit(lit):
        if isinstance(lit, ba.NOT):
            return (lit.args[0].obj, 0)
        else:
            return (lit.obj, 1)
    def out(data):
        print(data, file=outfd)
    b2a = BoolToAN(ba, ls_of_lit, out)

    def relabel_condition(condition):
        def relabel(m) :
            name = m.group(1)
            if name not in ["and", "or", "not"]:
                return tr(name)
            return name
        return re.sub(r"\b(\w+)\b", relabel, condition)

    transitions = bcx.getElementsByTagName("transition")

    bar = None
    if CFG["console"]:
        try:
            from progress.bar import Bar
            bar = Bar("converting", max=len(transitions))
        except ImportError:
            pass

    done = {}
    non_frontier = set()

    acoop = "__pint_coop%d"
    ncoop = 0

    for t in transitions:
        orig = tr(t.getAttribute("ori"))
        target = tr(t.getAttribute("ext"))
        condition = relabel_condition(t.getAttribute("condition")) or "1"

        non_frontier.add(target)

        lorig, ltarget = ba.symbols(orig, target)

        choices = []
        condition = ba.dnf(ba.parse(condition))
        if condition == ba.TRUE:
            expr_up = lorig
        else:
            if isinstance(condition, ba.OR):
                choices = [node for node in condition.args \
                    if isinstance(node, ba.AND)]
            expr_up = condition & lorig
            expr_up = expr_up.subs({ltarget: ba.FALSE})
            expr_up = ba.dnf(expr_up)

        # check duplicates
        if target in done:
            isdup = False
            for e in done[target]:
                if e == expr_up:
                    isdup = True
                    break
            if isdup:
                if bar:
                    bar.next()
                continue
            done[target].append(expr_up)
        else:
            done[target] = [expr_up]

        b2a.make_transitions([(target, 0, 1)], expr_up)

        if condition != ba.TRUE:
            condition = condition.subs({ltarget: ba.TRUE}).simplify()

        if condition == ba.TRUE:
            expr_down = ~lorig
        else:
            if len(choices) >= relax_threshold:
                for clause in choices:
                    ncoop += 1
                    a = acoop % ncoop
                    declare_automaton(a, 0, 1)
                    b2a.make_transitions([(a, 0, 1)], clause)
                    b2a.make_transitions([(a, 1, 0)], ~clause)
                    (la,) = ba.symbols(a)
                    condition = condition.subs({clause:la})

            expr_down = ~condition | ~lorig

        b2a.make_transitions([(target, 1, 0)], expr_down)

        if bar:
            bar.next()
    if bar:
        bar.finish()

    init_procs = ["%s=1" % a for a in PermNodes]
    if full_context:
        # add frontiers
        nodes = set(SimpleNodes)
        nodes.difference_update(non_frontier)
        init_procs += ["%(sort)s=0, %(sort)s=1" %{"sort": n} for n in nodes]
        bar = None
    print("initial_context %s" % ", ".join(init_procs), file=outfd)





import re

pat_pred = re.compile("(\w+)\((.+)\)\.")
pat_reaction = re.compile("(?:(\w+):)?(?:(.*)\s+for\s+)?(.*[^<])\s*(<?=>)\s*(.*)\.")
pat_sto = re.compile("^(?:\d\*)?\(?([^\d\(\)][^\(\)]*)\)?$")

class model:
    def __init__(self):
        self.present = set()
        self.molecules = set()
        self.reactions = {}

    def new_anon_reaction():
        n = len(self.reactions)
        return "_anon%d" % n

    def register_reaction(self, name, reactants, products):
        assert (name not in self.reactions)
        self.molecules.update(reactants)
        self.molecules.update(products)
        self.reactions[name] = (reactants, products)

    @classmethod
    def from_file(celf, filename):

        def parse_hs(hs):
            def fmt_sto(a):
                a = a.strip()
                p = pat_sto.match(a)
                assert (p), (a,p)
                a = p.group(1)
                return a
            return [fmt_sto(a) for a in hs.split("+") if a != "_"]

        m = celf()


        with open(filename) as f:
            for l in f:
                l = l.strip()
                if not l or l.startswith("%"):
                    continue
                mp = pat_pred.match(l)
                if mp:
                    pred = mp.group(1)
                    args = [a.strip() for a in mp.group(2).split(",")]
                    if pred == "present":
                        m.present.add(args[0])
                    elif pred == "absent":
                        try:
                            m.present.remove(args[0])
                        except KeyError:
                            pass
                else:
                    mr = pat_reaction.match(l)
                    if mr:
                        name, kinetics, reactants, direction, products = mr.groups()
                        if name is None:
                            name = m.new_anon_reaction()
                        reactants = parse_hs(reactants)
                        products = parse_hs(products)
                        if direction == "<=>":
                            if kinetics:
                                assert (kinetics[0]=="(" and kinetics[-1]==")")
                                kinetics = [k.strip() for k in kinetics[1:-1].split(",")]
                            if not kinetics or kinetics[0] != "0":
                                m.register_reaction(name, reactants, products)
                            if not kinetics or kinetics[1] != "0":
                                m.register_reaction("%s_rev" % name, products, reactants)
                        else:
                            if kinetics:
                                kinetics = kinetics.strip()
                            if not kinetics or kinetics != "0":
                                m.register_reaction(name, reactants, products)

        return m

    def an_name(self, molecule):
        return '"%s"' % molecule.replace("::","__")

    def to_an(self, fd):
        n = self.an_name

        def fmt_cond(conds):
            s = " and ".join(["%s=1" % n(c) for c in conds])
            if s:
                return " when %s" % s
            return ""

        for a in self.molecules:
            print("%s [0,1]" % n(a), file=fd)
        for e, (conds, res) in sorted(self.reactions.items()):
            print("%s [0,1,2]" % n(e), file=fd)
            print("%s 0 -> 1%s" % (n(e), fmt_cond(conds)), file=fd)
            print("%s 1 -> 2%s" % (n(e), fmt_cond(res)), file=fd)
            print("%s 2 -> 0" % n(e), file=fd)
            for a in conds:
                print("%s 1 -> 0 when %s=2" % (n(a), n(e)), file=fd)
            for a in res:
                print("%s 0 -> 1 when %s=1" % (n(a), n(e)), file=fd)

        init = ", ".join(["%s=1" % n(a) for a in self.present])
        print("initial_state %s" % init, file=fd)

def import_biocham(localfile, outfd):
    m = model.from_file(localfile)
    m.to_an(outfd)




#!/usr/bin/env python
# -*- coding: utf-8

import os
import re
import sys

if sys.version_info[0] < 3:
    raise Exception("python >= 3 is required")

from setuptools import setup, find_packages
from setuptools.command.install import install

NAME = "pypint"

META = {}
META_FILE = os.path.join(NAME, "__init__.py")
with open(META_FILE) as f:
    __data = f.read()
for key in ["version"]:
    match = re.search(r"^__{0}__ = ['\"]([^'\"]*)['\"]".format(key), __data, re.M)
    if not match:
        raise RuntimeError("Unable to find __{meta}__ string.".format(meta=key))
    META[key] = match.group(1)

setup(name = NAME,
    author = "Loïc Paulevé",
    author_email = "loic.pauleve@ens-cachan.org",
    url = "http://loicpauleve.name/pint",
    license = "CeCILL",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords='',
    description = "Python interface to Pint",
    long_description=open('README.rst').read(),
    python_requires=">=3.4",
    install_requires = [
        "colomoto_jupyter[networkx] >=0.4.1, <0.5",
        "networkx >= 2.0",
    ],
    extras_require = {
        "cadbiom": ["boolean.py"],
    },
    entry_points = {
        "console_scripts": [
            "pint-import=pypint.console:import_model",
        ],
    },
    include_package_data = True,
    packages = find_packages(exclude=["tests"]),
    **META
)

